--Program to show the differences between two text files.
--Author: R. M. Forno - Version 1.1 - 2002/07/15.

--Syntax: ex diff11 <file1_name> <file2_name> [<option>] [<max_time>]

--Output goes to standard output, so it can be redirected to a file.

--<option> may be omitted; its default value is 0. In this case, the
-- output consists of the lines from file1 that do not match file2 lines,
-- preceded by the > sign, then the lines from file2 that do not match
-- file1 lines, preceded by the < sign, and then the matching lines without
-- a preceding sign. This starts again when no more matching lines remain.
--If <option> is 0, then the output will be as described above.
--If <option> is 1, then only the number of the lines will be shown,
-- instead of the lines themselves.
--If <option> is 2, then the matching lines will not be shown, but the
-- non-matching lines will appear entirely.
--If <option> is 3, then the matching lines will not be shown, and only
-- the numbers of the non-matching lines will appear, preceded either
-- by the > or < sign.
--Other values of <option> are invalid.

--<max_time> may be omitted. In this case, a backtracking algorithm at the
-- end of processing is not executed.
--If <max_time> is 0, then this backtracking algorithm will be executed
-- with no time limit previously set, until all possibilities are scanned.
--If <max_time> is not 0, then this backtracking algorithm will be executed
-- up to max_time seconds.
--It is recommended either not to use this parameter or set it to a non-zero
-- value, because if not, in some cases the program might run nearly forever.

--The matching of lines is done on the basis of strict equality.
-- But the user has the option of changing this by providing his/her
-- own comparison function.
-- The include file compare.e contains the function custom_compare,
-- that merely calls the standard compare function. The user may
-- provide a different custom_compare function, for example to
-- disregard heading blanks, or case of letters, etc. The only
-- place to make this modification is the compare.e include file.

--This Version 1.1 uses an algorithm that is fast is no backtrack is
-- specified, but that not insures an optimal result in the sense of
-- minimizing the number of differences. On the other hand, if backtrack
-- is specified with no time limit, an optimal result is guaranteed, at the
-- cost of running nearly forever in many cases.

include get.e
include compare.e

sequence s, t, s1, t1

function GradeUp1(sequence s, integer n)
--This function is a merge sort of the Von Neumann kind, that is, 
-- first generates variable length sequences and then merges them.
--While it is more complex than the merge sort provided in EUPHORIA
-- demo ALLSORTS, it has some advantages:
--1) According to the argument n being zero or not zero, it respectively
-- gives the indexes of the sorted elements of s or the elements themselves.
-- The sorted indexes are useful when one sorts by a key but wants
-- to sort the remaining data, that for example resides in a "parallel"
-- sequence. You can see an example of that use in the IndexOf routine
-- that follows.
--2) This sort is faster than most sorts in demo ALLSORTS for RANDOM data.
-- Its performance is approximately equal to the stardard sort provided
-- with EUPHORIA (Shell sort). However, its main advantage is being MUCH
-- FASTER than most sorts for data that is nearly ordered, either in NORMAL
-- or REVERSE order, or in large chunks of data ordered both ways.
--3) It is stable, that is, it preserves the original order between
-- equal elements. Some of the other sorts are not stable (quick_sort,
-- for example). This is not important if the output consists in the
-- elements themselves, but it matters if the output consists in the
-- indexes.
--This is a modification of GradeUp from genfunc.e for use with a
-- custom_compare function.
    sequence z, r, w, t
    object low, high, temp, one, two
    integer len, m, k, g, h, f, a, b, mid, indz, indr, indt, indw, len1
    len = length(s)
    if len <= 1 then
	if n then
	    return s
	else
	    return repeat(1, len)
	end if
    end if
    indz = 1
    z = repeat(0, len)
    mid = len - 1
    indr = 0
    r = repeat(0, mid) --Indexes of developed data sequences
    t = repeat(0, len + mid) --Work area for developing sequences
    m = 1
    while m <= len do --Develop ordered data sequences from the input
	indr += 1
	low = s[m]
	high = low
	t[len] = m
	a = len
	b = len
	while m < len do
	    m += 1
	    temp = s[m]
	    if custom_compare(temp, high) >= 0 then --Data in normal order
		b += 1
		t[b] = m
		high = temp
	    elsif custom_compare(temp, low) < 0 then --Data in reverse order
		a -= 1
		t[a] = m
		low = temp
	    else --End of sequenced data
		m -= 1
		exit
	    end if
	end while
	z[indz..m] = t[a..b] --Add sequence to result
	m += 1
	indz = m
	r[indr] = m
    end while
    if indr <= 1 then --If only one data sequence developed
	if n then
	    for i = 1 to len do
		z[i] = s[z[i]]
	    end for
	end if
	return z
    end if
    len1 = len + 1
    r = r[1..indr]
    t = r
    w = z
    while 1 do --Start merge process
	indw = 0
	indt = 0
	k = 1 --First bound
	for i = 2 to indr by 2 do --Merge pairs of data sequences
	    g = r[i - 1]
	    h = r[i]
	    f = g
	    a = z[k]
	    one = s[a]
	    b = z[g]
	    two = s[b]
	    while 1 do --Merge two data sequences
		indw += 1
		if custom_compare(one, two) <= 0 then
		    w[indw] = a
		    k += 1
		    if k >= f then
			exit
		    end if
		    a = z[k]
		    one = s[a]
		else
		    w[indw] = b
		    g += 1
		    if g >= h then
			exit
		    end if
		    b = z[g]
		    two = s[b]
		end if
	    end while
	    --Both conditions cannot hold true at the same time
	    if k < f then
		indw += 1
		f -= 1
		b = indw + f - k
		w[indw..b] = z[k..f] --Remainder of first data sequence
		indw = h - 1
	    elsif g < h then
		indw = h - 1
		w[g..indw] = z[g..indw] --Remainder of second data sequence
	    end if
	    indt += 1
	    t[indt] = h
	    k = h --First bound = previous last bound
	end for
	if and_bits(1, indr) then --Orphaned last data sequence
	    w[k..len] = z[k..len]
	    indt += 1
	    t[indt] = len1
	end if
	if indt <= 1 then --If only one data sequence remains
	    if n then
		for i = 1 to len do
		    w[i] = s[w[i]]
		end for
	    end if
	    return w
	end if
	r = t[1..indt]
	indr = indt
	z = w
    end while
end function

procedure GenInd()
--Generate two sequences s1 & t1 mimicking the original sequences s & t
-- with respect to comparisons for equality, but consisting only of
-- integers, in order to diminish processing time.
    sequence x, ind, r, p, z
    integer len, lens, lent, k, d
    lens = length(s)
    lent = length(t)
    len = lens + lent
    if len = 0 then
	s1 = {}
	t1 = {}
	return
    end if
    x = s & t --Form composite sequence
    ind = GradeUp1(x, 0) --Generate index vector
    p = repeat(0, len)
    k = 1
    d = ind[1]
    r = x[d]
    p[d] = k
    for i = 2 to len do --Look for repeated sub-sequences
	d = ind[i]
	z = x[d]
	if not equal(z, r) then
	    k += 1
	    r = z
	end if
	p[d] = k
    end for
    s1 = p[1..lens]
    t1 = p[lens + 1..len]
end procedure

integer count, best
sequence count_match, best_match

procedure Dif(sequence s, sequence t)
--Generate a nearly-optimum try. count is the number of matching
-- pairs and count_match is the sequence of the respective index pairs.
    integer i, lens, lent, max, min, a, b, z, n, m, topi, x
    lens = length(s)
    lent = length(t)
    max = lens + lent
    count = 0
    count_match = {}
    i = 1
    z = 1
    while i <= lens do
	min = max
	topi = lens + 1
	m = i
	while m < topi do
	    x = s[m] --Speedwise
	    n = min - m + i + z - 1
	    if n > lent then
		n = lent
	    end if
	    for j = z to n do --Search for minimum slack
		if x = t[j] then
		    min = m - i + j - z --Update minimum slack
		    if min + i < topi then
			topi = min + i --Update upper limit for m
		    end if
		    a = m
		    b = j
		    exit
		end if
	    end for
	    m += 1
	end while
	if min < max then
	    count += 1
	    count_match = append(count_match, {a, b})
	    i = a + 1 --Update starting points
	    z = b + 1
	else
	    exit
	end if
    end while
end procedure

procedure Dif1(sequence s, sequence t)
--Generate a matching-pair list using the first match.
    integer lens, lent, i, z, m, j, x
    lens = length(s)
    lent = length(t)
    count = 0
    count_match = {}
    z = 1
    i = 1
    while i <= lens do
	m = i
	while m <= lens do
	    x = s[m] --Speedwise
	    j = z
	    while j <= lent do
		if x = t[j] then
		    count += 1
		    count_match = append(count_match, {m, j})
		    exit
		end if
		j += 1
	    end while
	    if j <= lent then --If exited, then exit
		exit
	    end if
	    m += 1
	end while
	if m <= lens then --If exited
	    i = m + 1 --Update starting points
	    z = j + 1
	else
	    exit
	end if
    end while
end procedure

atom t0, time_limit

procedure Diff(sequence s1, sequence t1, integer s2,
	integer t2, sequence res, integer same)
--Backtracking algorithm.
    integer lens, lent, n, m, k
    lens = length(s1)
    lent = length(t1)
    if time_limit then
	if time() - t0 > time_limit then
	    return
	end if
    end if
    if same + lent - t2 <= best then
	return
    end if
    t2 += 1
    n = lens - best + same
    for i = s2 + 1 to n do
	k = s1[i]
	m = lent - best + same
	for j = t2 to m do
	    if k = t1[j] then
		same += 1
		res = append(res, {i, j})
		if same > best then
		    best = same
		    best_match = res
		end if
		Diff(s1, t1, i, j, res, same)
		same -= 1
		res = res[1..length(res) - 1]
	    end if
	end for
    end for
end procedure

function ReadIn(sequence fn)
--Read-in input files as sequences.
    sequence s
    object x
    integer f
    s = {}
    f = open(fn, "r")
    if f < 0 then
	puts(2, "Error - cannot open " & fn)
	abort(3)
    end if
    while 1 do
	x = gets(f)
	if atom(x) then
	    exit
	end if
	s = append(s, x)
    end while
    return s
end function

function OutOption(integer j, integer option, sequence s)
    if option = 0 or option = 2 then
	return s[j]
    end if
    return sprintf("%d\n", j)
end function

procedure OutDiff(sequence s, sequence t, sequence best_match, integer option)
--Ouput resulting in, out, and equal lines.
    sequence x
    integer a, b, u, v
    a = 1
    b = 1
    for i = 1 to length(best_match) do
	x = best_match[i]
	u = x[1]
	v = x[2]
	for j = a to u - 1 do
	    puts(1, ">" & OutOption(j, option, s))
	end for
	for j = b to v - 1 do
	    puts(1, "<" & OutOption(j, option, t))
	end for
	if option <= 1 then
	    puts(1, OutOption(u, option, s))
	end if
	a = u + 1
	b = v + 1
    end for
    for j = a to length(s) do
	puts(1, ">" & OutOption(j, option, s))
    end for
    for j = b to length(t) do
	puts(1, "<" & OutOption(j, option, t))
    end for
end procedure

function Invert(sequence s)
--Invert sequence of matching pairs for swapped files.
    sequence x
    for i = 1 to length(s) do
	x = s[i]
	s[i] = {x[2], x[1]}
    end for
    return s
end function

procedure Difference()
--Main procedure. It calls first a non-backtracking algorithm to find a first
-- best result, then the same with swapped files, then a first-try algorithm,
-- then the same with swapped files, and last (conditionally) a
-- backtracking algorithm, each time trying to improve the best result up to
-- this point.
    sequence cmd, x, z
    integer len, backtrack
    object option
    cmd = command_line()
    len = length(cmd)
    if len < 4 then
	puts(2, "Error. Command line should have at least 2 arguments\n")
	puts(2, "Syntax: ex dif <1st file> <2nd file> [<option>] [<time]\n")
	abort(1)
    end if
    option = '0' --Default
    backtrack = 0 --Default
    if len >= 5 then
	option = cmd[5]
	if length(option) != 1 then
	    puts(2, "Incorrect option " & option & '\n')
	    puts(2, "Allowable values: 0 1 2 3\n")
	    abort(2)
	end if
	option = option[1]
	if not find(option, "0123") then
	    puts(2, "Incorrect option " & option & '\n')
	    puts(2, "Allowable values: 0 1 2 3\n")
	    abort(2)
	end if
    end if
    option -= '0'
    if len >= 6 then
	backtrack = 1
	z = cmd[6]
	x = value(z)
	if x[1] = GET_SUCCESS then
	    time_limit = x[2]
	else
	    puts(2, "Error - 4th argument not a valid number: " & z)
	    abort(2)
	end if
    end if
    s = ReadIn(cmd[3])
    t = ReadIn(cmd[4])
    GenInd() --To decrease processing time
    Dif(s1, t1) --Non-backtracking algorithm
    best = count
    best_match = count_match
    Dif(t1, s1) --Same with swapped files
    if count > best then
	best = count
	best_match = Invert(count_match)
    end if
    Dif1(s1, t1) --First-try algorithm
    if count > best then
	best = count
	best_match = count_match
    end if
    Dif1(t1, s1) --Same with swapped files
    if count > best then
	best = count
	best_match = Invert(count_match)
    end if
    if backtrack then
	t0 = time()
	Diff(s1, t1, 0, 0, {}, 0) --Backtracking algorithm
    end if
    OutDiff(s, t, best_match, option) --Output result
end procedure

Difference()
------END OF FILE------
