--Program to show the differences between two text files.
--Author: R. M. Forno - Version 5.0 - 2002/07/15.

--Syntax: ex diff50 <file1_name> <file2_name> [<option>] [<max_time>]

--Output goes to standard output, so it can be redirected to a file.

--<option> may be omitted; its default value is 0. In this case, the
-- output consists of the lines from file1 that do not match file2 lines,
-- preceded by the > sign, then the lines from file2 that do not match
-- file1 lines, preceded by the < sign, and then the matching lines without
-- a preceding sign. This starts again when no more matching lines remain.
--If <option> is 0, then the output will be as described above.
--If <option> is 1, then only the number of the lines will be shown,
-- instead of the lines themselves.
--If <option> is 2, then the matching lines will not be shown, but the
-- non-matching lines will appear entirely.
--If <option> is 3, then the matching lines will not be shown, and only
-- the numbers of the non-matching lines will appear, preceded either
-- by the > or < sign.
--Other values of <option> are invalid.

--<max_time> may be omitted. In this case, a backtracking algorithm at the
-- end of processing is not executed.
--If <max_time> is 0, then this backtracking algorithm will be executed
-- with no time limit previously set, until all possibilities are scanned.
--If <max_time> is not 0, then this backtracking algorithm will be executed
-- up to max_time seconds.
--It is recommended either not to use this parameter or set it to a non-zero
-- value, because if not, in some cases the program might run nearly forever.

--This program does not guarantee that an optimal output will be generated
-- (optimal in the sense of maximizing the number of matched lines),
-- but this optimal will be attained in about 99% of the practical cases.
-- The sub-optimal solutions will be usually very good, too.
-- Version 1 guarantees an optimal solution, at the price of
-- running nearly forever in many cases, when using the backtracking
-- algorithm option.
-- This version, rather than maximize the number of matched lines, tries
-- that the matching groups have maximum length, which is more practical
-- for the user.
-- This version is usually slower than Version 1 when no backtracking is
-- specified in both versions, but faster if backtracking is specified.

--The matching of lines is done on the basis of strict equality.
-- But the user has the option of changing this by providing his/her
-- own comparison function.
-- The include file compare.e contains the function custom_compare,
-- that merely calls the standard compare function. The user may
-- provide a different custom_compare function, for example to
-- disregard heading blanks, or case of letters, etc. The only
-- place to make this modification is the compare.e include file.

--This Version 5.0 uses an algorithm that is a mix of versions 1.1 and 2.0.

include get.e
include compare.e

sequence s, t, s1, t1, s2, t2, s3, t3, sl, tl

function GradeUp1(sequence s, integer n)
--This function is a merge sort of the Von Neumann kind, that is, 
-- first generates variable length sequences and then merges them.
--While it is more complex than the merge sort provided in EUPHORIA
-- demo ALLSORTS, it has some advantages:
--1) According to the argument n being zero or not zero, it respectively
-- gives the indexes of the sorted elements of s or the elements themselves.
-- The sorted indexes are useful when one sorts by a key but wants
-- to sort the remaining data, that for example resides in a "parallel"
-- sequence. You can see an example of that use in the IndexOf routine
-- that follows.
--2) This sort is faster than most sorts in demo ALLSORTS for RANDOM data.
-- Its performance is approximately equal to the stardard sort provided
-- with EUPHORIA (Shell sort). However, its main advantage is being MUCH
-- FASTER than most sorts for data that is nearly ordered, either in NORMAL
-- or REVERSE order, or in large chunks of data ordered both ways.
--3) It is stable, that is, it preserves the original order between
-- equal elements. Some of the other sorts are not stable (quick_sort,
-- for example). This is not important if the output consists in the
-- elements themselves, but it matters if the output consists in the
-- indexes.
--This is a modification of GradeUp from genfunc.e for use with a
-- custom_compare function.
    sequence z, r, w, t
    object low, high, temp, one, two
    integer len, m, k, g, h, f, a, b, mid, indz, indr, indt, indw, len1
    len = length(s)
    if len <= 1 then
	if n then
	    return s
	else
	    return repeat(1, len)
	end if
    end if
    indz = 1
    z = repeat(0, len)
    mid = len - 1
    indr = 0
    r = repeat(0, mid) --Indexes of developed data sequences
    t = repeat(0, len + mid) --Work area for developing sequences
    m = 1
    while m <= len do --Develop ordered data sequences from the input
	indr += 1
	low = s[m]
	high = low
	t[len] = m
	a = len
	b = len
	while m < len do
	    m += 1
	    temp = s[m]
	    if custom_compare(temp, high) >= 0 then --Data in normal order
		b += 1
		t[b] = m
		high = temp
	    elsif custom_compare(temp, low) < 0 then --Data in reverse order
		a -= 1
		t[a] = m
		low = temp
	    else --End of sequenced data
		m -= 1
		exit
	    end if
	end while
	z[indz..m] = t[a..b] --Add sequence to result
	m += 1
	indz = m
	r[indr] = m
    end while
    if indr <= 1 then --If only one data sequence developed
	if n then
	    for i = 1 to len do
		z[i] = s[z[i]]
	    end for
	end if
	return z
    end if
    len1 = len + 1
    r = r[1..indr]
    t = r
    w = z
    while 1 do --Start merge process
	indw = 0
	indt = 0
	k = 1 --First bound
	for i = 2 to indr by 2 do --Merge pairs of data sequences
	    g = r[i - 1]
	    h = r[i]
	    f = g
	    a = z[k]
	    one = s[a]
	    b = z[g]
	    two = s[b]
	    while 1 do --Merge two data sequences
		indw += 1
		if custom_compare(one, two) <= 0 then
		    w[indw] = a
		    k += 1
		    if k >= f then
			exit
		    end if
		    a = z[k]
		    one = s[a]
		else
		    w[indw] = b
		    g += 1
		    if g >= h then
			exit
		    end if
		    b = z[g]
		    two = s[b]
		end if
	    end while
	    --Both conditions cannot hold true at the same time
	    if k < f then
		indw += 1
		f -= 1
		b = indw + f - k
		w[indw..b] = z[k..f] --Remainder of first data sequence
		indw = h - 1
	    elsif g < h then
		indw = h - 1
		w[g..indw] = z[g..indw] --Remainder of second data sequence
	    end if
	    indt += 1
	    t[indt] = h
	    k = h --First bound = previous last bound
	end for
	if and_bits(1, indr) then --Orphaned last data sequence
	    w[k..len] = z[k..len]
	    indt += 1
	    t[indt] = len1
	end if
	if indt <= 1 then --If only one data sequence remains
	    if n then
		for i = 1 to len do
		    w[i] = s[w[i]]
		end for
	    end if
	    return w
	end if
	r = t[1..indt]
	indr = indt
	z = w
    end while
end function

function GradeUp(sequence s, integer n)
--This function is a merge sort of the Von Neumann kind, that is, 
-- first generates variable length sequences and then merges them.
--While it is more complex than the merge sort provided in EUPHORIA
-- demo ALLSORTS, it has some advantages:
--1) According to the argument n being zero or not zero, it respectively
-- gives the indexes of the sorted elements of s or the elements themselves.
-- The sorted indexes are useful when one sorts by a key but wants
-- to sort the remaining data, that for example resides in a "parallel"
-- sequence. You can see an example of that use in the IndexOf routine
-- that follows.
--2) This sort is faster than most sorts in demo ALLSORTS for RANDOM data.
-- Its performance is approximately equal to the stardard sort provided
-- with EUPHORIA (Shell sort). However, its main advantage is being MUCH
-- FASTER than most sorts for data that is nearly ordered, either in NORMAL
-- or REVERSE order, or in large chunks of data ordered both ways.
--3) It is stable, that is, it preserves the original order between
-- equal elements. Some of the other sorts are not stable (quick_sort,
-- for example). This is not important if the output consists in the
-- elements themselves, but it matters if the output consists in the
-- indexes.
--This is the original GradeUp function from genfunc.e.
    sequence z, r, w, t
    object low, high, temp, one, two
    integer len, m, k, g, h, f, a, b, mid, indz, indr, indt, indw, len1
    len = length(s)
    if len <= 1 then
	if n then
	    return s
	else
	    return repeat(1, len)
	end if
    end if
    indz = 1
    z = repeat(0, len)
    mid = len - 1
    indr = 0
    r = repeat(0, mid) --Indexes of developed data sequences
    t = repeat(0, len + mid) --Work area for developing sequences
    m = 1
    while m <= len do --Develop ordered data sequences from the input
	indr += 1
	low = s[m]
	high = low
	t[len] = m
	a = len
	b = len
	while m < len do
	    m += 1
	    temp = s[m]
	    if compare(temp, high) >= 0 then --Data in normal order
		b += 1
		t[b] = m
		high = temp
	    elsif compare(temp, low) < 0 then --Data in reverse order
		a -= 1
		t[a] = m
		low = temp
	    else --End of sequenced data
		m -= 1
		exit
	    end if
	end while
	z[indz..m] = t[a..b] --Add sequence to result
	m += 1
	indz = m
	r[indr] = m
    end while
    if indr <= 1 then --If only one data sequence developed
	if n then
	    for i = 1 to len do
		z[i] = s[z[i]]
	    end for
	end if
	return z
    end if
    len1 = len + 1
    r = r[1..indr]
    t = r
    w = z
    while 1 do --Start merge process
	indw = 0
	indt = 0
	k = 1 --First bound
	for i = 2 to indr by 2 do --Merge pairs of data sequences
	    g = r[i - 1]
	    h = r[i]
	    f = g
	    a = z[k]
	    one = s[a]
	    b = z[g]
	    two = s[b]
	    while 1 do --Merge two data sequences
		indw += 1
		if compare(one, two) <= 0 then
		    w[indw] = a
		    k += 1
		    if k >= f then
			exit
		    end if
		    a = z[k]
		    one = s[a]
		else
		    w[indw] = b
		    g += 1
		    if g >= h then
			exit
		    end if
		    b = z[g]
		    two = s[b]
		end if
	    end while
	    --Both conditions cannot hold true at the same time
	    if k < f then
		indw += 1
		f -= 1
		b = indw + f - k
		w[indw..b] = z[k..f] --Remainder of first data sequence
		indw = h - 1
	    elsif g < h then
		indw = h - 1
		w[g..indw] = z[g..indw] --Remainder of second data sequence
	    end if
	    indt += 1
	    t[indt] = h
	    k = h --First bound = previous last bound
	end for
	if and_bits(1, indr) then --Orphaned last data sequence
	    w[k..len] = z[k..len]
	    indt += 1
	    t[indt] = len1
	end if
	if indt <= 1 then --If only one data sequence remains
	    if n then
		for i = 1 to len do
		    w[i] = s[w[i]]
		end for
	    end if
	    return w
	end if
	r = t[1..indt]
	indr = indt
	z = w
    end while
end function

function GenInd1(sequence s, sequence t)
--Generate a sequence of two sequences mimicking the original sequences s & t
-- with respect to comparisons for equality, but consisting only of
-- integers, in order to diminish processing time.
--This variant uses custom_compare.
    sequence x, ind, r, p, z
    integer len, lens, lent, k, d
    lens = length(s)
    lent = length(t)
    len = lens + lent
    if len = 0 then
	return {}
    end if
    x = s & t --Form composite sequence
    ind = GradeUp1(x, 0) --Generate index vector
    p = repeat(0, len)
    k = 1
    d = ind[1]
    r = x[d]
    p[d] = k
    for i = 2 to len do --Look for repeated sub-sequences
	d = ind[i]
	z = x[d]
	if custom_compare(z, r) != 0 then
	    k += 1
	    r = z
	end if
	p[d] = k
    end for
    return {p[1..lens], p[lens + 1..len]}
end function

function GenInd(sequence s, sequence t)
--Generate a sequence of two sequences mimicking the original sequences s & t
-- with respect to comparisons for equality, but consisting only of
-- integers, in order to diminish processing time.
--This variant uses standard compare.
    sequence x, ind, r, p, z
    integer len, lens, lent, k, d
    lens = length(s)
    lent = length(t)
    len = lens + lent
    if len = 0 then
	return {}
    end if
    x = s & t --Form composite sequence
    ind = GradeUp(x, 0) --Generate index vector
    p = repeat(0, len)
    k = 1
    d = ind[1]
    r = x[d]
    p[d] = k
    for i = 2 to len do --Look for repeated sub-sequences
	d = ind[i]
	z = x[d]
	if not equal(z, r) then
	    k += 1
	    r = z
	end if
	p[d] = k
    end for
    return {p[1..lens], p[lens + 1..len]}
end function

function MaxMatchGroup(sequence s1, sequence t1)
--Search for the maximum length grouping of elements.
    sequence old, new, best
    integer lens, lent, lenb, k, ns, nt
    object a, b
    lens = length(s1)
    lent = length(t1)
    lenb = 0
    best = {lenb, 0, 0} --Initialize best grouping
    new = {} --To avoid searching when no improvement is possible
    for i = 1 to lens do
	a = s1[i]
	old = new
	new = {}
	if atom(a) then --Disregard previous groupings
	    for j = 1 to lent do
		b = t1[j]
		if equal(a, b) then
		    if j < lent then
			if i < lens then
			    if equal(s1[i + 1], t1[j + 1]) then
				new &= j --Prepare already-done test
			    end if
			end if
		    end if
		    if not find(j - 1, old) then --Not previously done
			k = 1
			while 1 do --Determine group
			    ns = i + k
			    if ns > lens then
				exit
			    end if
			    nt = j + k
			    if nt > lent then
				exit
			    end if
			    if not equal(s1[ns], t1[nt]) then
				exit
			    end if
			    k += 1
			end while
			if k > lenb then --Update best grouping
			    lenb = k
			    best = {lenb, i, j}
			end if
		    end if
		end if
	    end for
	end if
    end for
    return best
end function

function NonMatchGroup(sequence s)
--Group non-matching elements
    sequence r, p
    object a
    integer k, len
    len = length(s)
    k = 1
    r = {}
    while k <= len do
	a = s[k]
	if atom(a) then
	    p = {a}
	    k += 1
	    while k <= len do
		a = s[k]
		if atom(a) then
		    p &= a
		else
		    exit
		end if
		k += 1
	    end while
	    r = append(r, p)
	else
	    r = append(r, a)
	    k += 1
	end if
    end while
    return r
end function

function MatchGroup(sequence s1, sequence t1)
--Group elements in descending order of group lengths.
    integer len, a, b
    sequence best
    while 1 do --Search for groups
	best = MaxMatchGroup(s1, t1)
	len = best[1]
	if len = 0 then
	    exit
	end if
	a = best[2]
	b = best[3]
	s1 = s1[1..a - 1] & {s1[a..a + len - 1]} & s1[a + len..length(s1)]
	t1 = t1[1..b - 1] & {t1[b..b + len - 1]} & t1[b + len..length(t1)]
    end while
    --Group non-matching elements
    s1 = NonMatchGroup(s1)
    t1 = NonMatchGroup(t1)
    return {s1, t1}
end function

function GetLengths(sequence s)
--Returns the lengths of the subsequences in s.
    sequence r
    integer len
    len = length(s)
    r = repeat(0, len)
    for i = 1 to len do
	r[i] = length(s[i])
    end for
    return r
end function

function Dif(sequence s, sequence t, sequence sl)
--Generate a nearly-optimum try. count is the number of matching
-- pairs and count_match is the sequence of the respective index pairs.
    integer i, lens, lent, max, min, a, b, z, n, m, topi, x, count
    sequence count_match
    lens = length(s)
    lent = length(t)
    max = lens + lent
    count = 0
    count_match = {}
    i = 1
    z = 1
    while i <= lens do
	min = max
	topi = lens + 1
	m = i
	while m < topi do
	    x = s[m]
	    n = min - m + i + z - 1
	    if n > lent then
		n = lent
	    end if
	    for j = z to n do --Search for minimum slack
		if x = t[j] then
		    min = m - i + j - z --Update minimum slack
		    if min + i < topi then
			topi = min + i --Update upper limit for m
		    end if
		    a = m
		    b = j
		    exit
		end if
	    end for
	    m += 1
	end while
	if min < max then
	    count += sl[a]
	    count_match = append(count_match, {a, b})
	    i = a + 1 --Update starting points
	    z = b + 1
	else
	    exit
	end if
    end while
    return count & count_match
end function

function Dif1(sequence s, sequence t, sequence sl)
--Generate a matching-pair list using the first match.
    integer lens, lent, i, z, m, j, x, count
    sequence count_match
    lens = length(s)
    lent = length(t)
    count = 0
    count_match = {}
    z = 1
    i = 1
    while i <= lens do
	m = i
	while m <= lens do
	    x = s[m]
	    j = z
	    while j <= lent do
		if x = t[j] then
		    count += sl[m]
		    count_match = append(count_match, {m, j})
		    exit
		end if
		j += 1
	    end while
	    if j <= lent then --If exited, then exit
		exit
	    end if
	    m += 1
	end while
	if m <= lens then --If exited
	    i = m + 1 --Update starting points
	    z = j + 1
	else
	    exit
	end if
    end while
    return count & count_match
end function

atom t0, time_limit
sequence xs, xt, res, best_match
integer best_up_to, ls3, lt3, best_count

procedure GenAcum()
--Generate running totals to test if improvement possible for backtracking
    integer max, k
    sequence as, at
    max = 0
    for i = 1 to ls3 do --Get maximum value in s3, t3
	if s3[i] > max then
	    max = s3[i]
	end if
    end for
    for i = 1 to lt3 do
	if t3[i] > max then
	    max = t3[i]
	end if
    end for
    as = repeat(0, max)
    at = as
    for i = 1 to ls3 do --Count number of possible matches
	as[s3[i]] += 1
    end for
    for i = 1 to lt3 do
	at[t3[i]] += 1
    end for
    for i = 1 to max do --Set the minimum for both sequences
	--(cannot match more than exists at the opposite side)
	if as[i] < at[i] then
	    at[i] = as[i]
	else
	    as[i] = at[i]
	end if
    end for
    xs = repeat(0, ls3)
    xt = repeat(0, lt3)
    for i = ls3 to 1 by -1 do --Store lengths up to maximum number of matches
	k = s3[i]
	if as[k] then
	    xs[i] = sl[i]
	    as[k] -= 1
	end if
    end for
    for i = lt3 to 1 by -1 do
	k = t3[i]
	if at[k] then
	    xt[i] = tl[i]
	    at[k] -= 1
	end if
    end for
    xs = 0 & xs
    xt = 0 & xt
    for i = 1 to ls3 do --Compute running totals
	xs[i + 1] += xs[i]
    end for
    k = xs[ls3 + 1]
    for i = 1 to ls3 + 1 do
	xs[i] = k - xs[i]
    end for
    for i = 1 to lt3 do
	xt[i + 1] += xt[i]
    end for
    k = xt[lt3 + 1]
    for i = 1 to lt3 + 1 do
	xt[i] = k - xt[i]
    end for
end procedure

procedure Diff(integer s2, integer t2)
--Backtracking algorithm.
    integer k, p
    if time_limit then
	if time() - t0 > time_limit then
	    return
	end if
    end if
    t2 += 1
    s2 += 1
    if best_up_to + xs[s2] + xt[t2] <= best_count then
	return
    end if
    for i = s2 to ls3 do
	k = s3[i]
	p = sl[i]
	for j = t2 to lt3 do
	    if k = t3[j] then
		best_up_to += p
		res = append(res, {i, j})
		if best_up_to > best_count then
		    best_count = best_up_to
		    best_match = res
		end if
		Diff(i, j)
		best_up_to -= p
		res = res[1..length(res) - 1]
	    end if
	end for
    end for
end procedure

function ReadIn(sequence fn)
--Read-in input files as sequences.
    sequence s
    object x
    integer f
    s = {}
    f = open(fn, "r")
    if f < 0 then
	puts(2, "Error - cannot open " & fn)
	abort(3)
    end if
    while 1 do
	x = gets(f)
	if atom(x) then
	    exit
	end if
	s = append(s, x)
    end while
    close(f)
    return s
end function

function OutOption(integer j, integer option, sequence s)
    if option = 0 or option = 2 then
	return s[j]
    end if
    return sprintf("%d\n", j)
end function

procedure OutDiff(sequence best_pairs, integer option)
--Ouput resulting in, out, and equal lines.
    sequence x
    integer a, b, u, v, z, zs, zt
    a = 1
    b = 1
    zs = 1
    zt = 1
    for i = 1 to length(best_pairs) do
	x = best_pairs[i]
	u = x[1]
	v = x[2]
	z = 0
	for j = a to u - 1 do
	    z += sl[j]
	end for
	for j = zs to zs + z - 1 do
	    puts(1, ">" & OutOption(j, option, s))
	end for
	zs += z
	z = 0
	for j = b to v - 1 do
	    z += tl[j]
	end for
	for j = zt to zt + z - 1 do
	    puts(1, "<" & OutOption(j, option, t))
	end for
	zt += z
	z = sl[u]
	for j = zs to zs + z - 1 do
	    if option <= 1 then
		puts(1, OutOption(j, option, s))
	    end if
	end for
	zs += z
	zt += z
	a = u + 1
	b = v + 1
    end for
    for j = zs to length(s) do
	puts(1, ">" & OutOption(j, option, s))
    end for
    for j = zt to length(t) do
	puts(1, "<" & OutOption(j, option, t))
    end for
end procedure

function InvertPairs(sequence p)
    for i = 1 to length(p) do
	p[i] = {p[i][2], p[i][1]}
    end for 
    return p
end function

procedure Difference()
--Main procedure.
    sequence cmd, s3p, t3p, slp, tlp, x, z, p
    integer len, count, backtrack
    object option
    cmd = command_line()
    len = length(cmd)
    if len < 4 then
	puts(2, "Error. Command line should have at least 2 arguments\n")
	puts(2, "Syntax: ex dif <1st file> <2nd file> [<option>] [<time]\n")
	abort(1)
    end if
    option = '0' --Default
    backtrack = 0 --Default
    if len >= 5 then
	option = cmd[5]
	if length(option) != 1 then
	    puts(2, "Incorrect option " & option & '\n')
	    puts(2, "Allowable values: 0 1 2 3\n")
	    abort(2)
	end if
	option = option[1]
	if not find(option, "0123") then
	    puts(2, "Incorrect option " & option & '\n')
	    puts(2, "Allowable values: 0 1 2 3\n")
	    abort(2)
	end if
    end if
    option -= '0'
    if len >= 6 then
	backtrack = 1
	z = cmd[6]
	x = value(z)
	if x[1] = GET_SUCCESS then
	    time_limit = x[2]
	else
	    puts(2, "Error - 4th argument not a valid number: " & z)
	    abort(2)
	end if
    end if
    s = ReadIn(cmd[3])
    t = ReadIn(cmd[4])
    p = GenInd1(s, t) --To decrease processing time
    s1 = p[1]
    t1 = p[2]
    p = MatchGroup(s1, t1) --Groups matching elements
    s2 = p[1]
    t2 = p[2]
    p = GenInd(s2, t2) --Also, to decrease processing time
    s3 = p[1]
    t3 = p[2]
    ls3 = length(s3)
    lt3 = length(t3)
    sl = GetLengths(s2) --Lengths, to be added up for evaluating performance
    tl = GetLengths(t2)
    p = Dif(s3, t3, sl)
    best_count = p[1]
    best_match = p[2..length(p)]
    p = Dif1(s3, t3, sl)
    count = p[1]
    if count > best_count then
	best_count = count
	best_match = p[2..length(p)]
    end if
    p = MatchGroup(t1, s1) --Try the inverse
    s2 = p[2]
    t2 = p[1]
    p = GenInd(t2, s2)
    s3p = p[2]
    t3p = p[1]
    slp = GetLengths(s2)
    tlp = GetLengths(t2)
    p = Dif(t3p, s3p, tlp)
    count = p[1]
    if count > best_count then
	best_count = count
	best_match = InvertPairs(p[2..length(p)])
	sl = slp
	tl = tlp
	s3 = s3p
	t3 = t3p
	ls3 = length(s3)
	lt3 = length(t3)
    end if
    p = Dif1(t3p, s3p, tlp)
    count = p[1]
    if count > best_count then
	best_count = count
	best_match = InvertPairs(p[2..length(p)])
	sl = slp
	tl = tlp
	s3 = s3p
	t3 = t3p
	ls3 = length(s3)
	lt3 = length(t3)
    end if
    if backtrack then
	GenAcum()
	res = {}
	best_up_to = 0
	t0 = time()
	Diff(0, 0)
    end if
    OutDiff(best_match, option) --Output result
end procedure

Difference()
------END OF FILE------
